! Copyright 2019
!
! For a comprehensive list of the developers that contributed to these codes
! see the UK-AMOR website.
!
! This file is part of UKRmol-in (UKRmol+ suite).
!
!     UKRmol-in is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
!
!     UKRmol-in is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with  UKRmol-in (in source/COPYING). Alternatively, you can also visit
!     <https://www.gnu.org/licenses/>.
module maths
!
! Interfacing module for library mathematical routines
! Joanne Carr, February 2010
!

 USE precisn, ONLY : wp ! for specifying the kind of reals
 implicit none
 private

 public :: maths_erfc ! calculates the complement of the error function erf(x)
 public :: diag       ! matrix diagonalizer for psn
 public :: maths_matrix_multiply ! simple matrix multiplication
 public :: maths_tcoef ! simple matrix transposition
 public :: maths_cveca_plus_vecb ! constant*veca + vecb 
 public :: maths_mxm ! matrix multplication in swtrmo; should be merged with maths_matrix_multiply at some point?? 

 public :: maths_dmatrix_multiply_blas95
 public :: maths_zmatrix_multiply_blas95
 interface  maths_matrix_multiply_blas95
    module procedure  maths_dmatrix_multiply_blas95
    module procedure  maths_zmatrix_multiply_blas95
 end interface 

 

 contains
  
   function maths_erfc(x)
! Calculates the complement of the error function erf(x)
! Note that ifail is always set to 0 in s15erfc, i.e. 
! no failure reported on exit
 
   USE global_utils, ONLY : s15erfc ! the location of the actual function

   REAL(KIND=wp), INTENT(IN) :: x

   REAL(KIND=wp) :: maths_erfc
   INTEGER :: ifail

   maths_erfc = s15erfc(x,ifail)

   end function maths_erfc

!*==diag.spg  processed by SPAG 6.56Rc at 17:18 on  7 Apr 2010
      SUBROUTINE DIAG(MK,H,EIG,VEC,IPRNT)
      IMPLICIT NONE
!*********************************************************************
!
!     DIAG DIAGONALIZES THE DENSITY MATRIX BLOCKS. THIS USED A NAG
!     LIBRARY ROUTINE F02FAF TO PERFORM THE DIAGONALIZATION.
!     This routine has been replaced by the lapack double
!     precision routine dsyev (nv 16/5/02)
!
!     INPUT DATA :
!              H  LOWER TRIANGLE OF THE DENSITY MATRIX (stored as a
!                 linear array with MK*(MK+1)/2 elements)
!             MK  SIZE OF THE DENSITY MATRIX IS MK BY MK
!          IPRNT  SWITCH FOR DEBUG PRINTING
!
!     OUTPUT DATA :
!             EIG  EIGEN-ENERGIES OF THE DENSITY MATRIX
!             VEC  EIGEN-VECTORS OF THE DENSITY MATRIX
!
!     TO SAVE SPACE AND TO COMPLY WITH THE REQUIREMENTS OF THE NAG
!     ROUTINE, THE LOWER TRIANGLE OF THE DENSITY MATRIX IS COPIED
!     INTO THE ARRAY VEC BEFORE THE CALL. IT IS OVERWRITTEN ON OUTPUT
!     BY THE EIGEN-VECTORS.
!
!*********************************************************************
!
!*** Start of declarations rewritten by SPAG
!
! Dummy arguments
!
      INTEGER :: IPRNT, MK
      REAL(KIND=wp), DIMENSION(MK) :: EIG
      REAL(KIND=wp), DIMENSION(*) :: H
      REAL(KIND=wp), DIMENSION(MK,MK) :: VEC
      INTENT (IN) H, IPRNT
      INTENT (INOUT) VEC
!
! Local variables
!
      INTEGER :: I, IFAIL, J, K, LX, NDIM
      REAL(KIND=wp), DIMENSION(3*mk) :: X
!
!*** End of declarations rewritten by SPAG
!

 9000 FORMAT(/,10X,' DEBUG INFORMATION IN SUBROUTINE DIAG',//)
 9010 FORMAT(/,10X,' HALF TRIANGLE DIMENSION = ',I5,/)
 9020 FORMAT(/,10X,' H-MATRIX ',/)
 9030 FORMAT(2X,6F12.8)
 9040 FORMAT(/,10X,'VEC MATRIX ',/)
 9050 FORMAT(/,10X,'EIGENVALUES ',/)
 9060 FORMAT((/6(I3,F10.5)))
 9070 FORMAT(/,10X,'EIGENVECTORS ',/)
 9080 FORMAT(2X,I5,6F12.5,/,(7X,6F12.5))
 9999 FORMAT(/,10X,'DIAG : LIBRARY ROUTINE RETURNS IFAIL = ',I2,/)
 9888 FORMAT(/,10X,'ERROR IN DIAG : INPUT MATRIX DIMENSION MK = ',I5,/)
 
      IFAIL = 0
!
!---- COPY LOWER TRIANGLE INTO LOWER HALF OF VEC.
!
      K=0
      DO I=1, MK
         DO J=1, I
            K=K+1
            VEC(I,J)=H(K)
            VEC(J,I)=H(K)
         END DO
      END DO
!
!---- DEBUG CODE. CHECKS IF EVERYTHING IS STORED OK.
!
      IF(IPRNT.GT.1)THEN
         NDIM=(MK*(MK+1))/2
         WRITE(*,9000)
         WRITE(*,9010)NDIM
         WRITE(*,9020)
         WRITE(*,9030)(H(K),K=1,NDIM)
         WRITE(*,9040)
         DO I=1, MK
            WRITE(*,9030)(VEC(J,I),J=1,MK)
         END DO
      END IF
 
      lx=3*mk
 
      IF(mk.GT.0)THEN
 
         CALL dsyev('V','L',mk,vec,mk,eig,x,lx,ifail)
         IF(IFAIL.NE.0)THEN
            WRITE(*,9999)IFAIL
            STOP
         END IF

      ELSE
         WRITE(*,9888)MK
         STOP
      END IF
!
!---- DEBUG CODE PRINTS OUT EIGEN VALUES AND VECTORS.
!
      IF(IPRNT.GT.1)THEN
         WRITE(*,9050)
         WRITE(*,9060)(I,EIG(I),I=1,MK)
         WRITE(*,9070)
         DO I=1, MK
            WRITE(*,9080)I, (VEC(J,I),J=1,MK)
         END DO
      END IF
 
      RETURN
      END SUBROUTINE DIAG

      subroutine maths_matrix_multiply(m,n,k,a,lda,b,ldb,c,ldc,lunit,ierr)
! (m by n) matrix C = (m by k) matrix A * (k by n) matrix B
      
      integer :: m, n, k
      integer :: lda, ldb, ldc ! leading dimensions of A, B and C, respectively
                               ! (should be >= m, k, m otherwise it's an error).
                               ! In dgemm, the A array is declared with dimensions 
                               ! (lda,*) and equivalently for B and C.
      integer :: lunit ! unit for printing
      integer :: ierr ! error code
      real(kind=wp) :: a(lda,k)
      real(kind=wp) :: b(ldb,n)
      real(kind=wp) :: c(ldc,n)

      ierr = 0

! some error checking
      if ((m.eq.0) .or. (n.eq.0) .or. (k.eq.0)) then
         write(lunit,*) ' WARNING in maths_matrix_multiply:'
         write(lunit,*) ' dimensions m, n, k = ',m,n,k
         ierr = 1
         return
      end if

! call the Level 3 blas routine, assuming that we don't need to 
! transpose either input matrix (hence the two 'n's), and that we 
! simply want to calculate c = a*b  i.e. c = 1.0_wp*a*b + 0.0_wp*c
!
! dgemm will also do some error checking.  For example, it will stop 
! if any of m,n,k are < 0

      call dgemm('n','n',m,n,k,1.0_wp,a,lda,b,ldb,0.0_wp,c,ldc)

      end subroutine maths_matrix_multiply


      subroutine maths_dmatrix_multiply_blas95(a,b,c,transpose_A,transpose_B,coeff_A,coeff_C)
! (m by n) matrix C = (m by k) matrix A * (k by n) matrix B
      implicit none

      integer :: m, n, k
      integer :: lda, ldb, ldc ! leading dimensions of A, B and C, respectively
                               ! (should be >= m, k, m otherwise it's an error).
                               ! In dgemm, the A array is declared with dimensions 
                               ! (lda,*) and equivalently for B and C.
      integer :: ierr ! error code
      real(kind=wp) :: a(:,:)
      real(kind=wp) :: b(:,:)
      real(kind=wp) :: c(:,:)
      real(kind=wp), optional :: coeff_A,coeff_C
      logical, optional :: transpose_A, transpose_B
      real(kind=wp) :: coeff_A_local,coeff_C_local
      character(len=1) :: transpose_A_local, transpose_B_local


      intent(in) a,b, transpose_A, transpose_B,coeff_A,coeff_C
      intent(inout) c

! Default values
      transpose_A_local='n'
      transpose_B_local='n'
      coeff_A_local = 1.0_wp
      coeff_C_local = 0.0_wp
      k = size(A,1)
      lda = max(1,size(a,1))
      ldb = max(1,size(b,1))
      ldc = max(1,size(c,1))
      m = size(C,1)
      n = size(C,2)

      if (present(transpose_A)) then
        if (transpose_A) then
          transpose_A_local='y'
          k = size(A,2)
        endif
      endif

      if (present(transpose_B)) then
        if (transpose_B) then
          transpose_B_local='y'
        endif
      endif

      if (present(coeff_A)) then
        coeff_A_local = coeff_A
      endif

      if (present(coeff_C)) then
        coeff_C_local = coeff_C
      endif

      ierr = 0
     
! call the Level 3 blas routine, assuming that we don't need to 
! transpose either input matrix (hence the two 'n's), and that we 
! simply want to calculate c = a*b  i.e. c = 1.0_wp*a*b + 0.0_wp*c
!
! dgemm will also do some error checking.  For example, it will stop 
! if any of m,n,k are < 0

      call dgemm(transpose_A_local,transpose_B_local,m,n,k,coeff_A_local,a,lda,b,ldb,coeff_C_local,c,ldc)

      end subroutine maths_dmatrix_multiply_blas95


      subroutine maths_zmatrix_multiply_blas95(a,b,c,transpose_A,transpose_B,coeff_A,coeff_C)
! (m by n) matrix C = (m by k) matrix A * (k by n) matrix B
      implicit none

      integer :: m, n, k
      integer :: lda, ldb, ldc ! leading dimensions of A, B and C, respectively
                               ! (should be >= m, k, m otherwise it's an error).
                               ! In dgemm, the A array is declared with dimensions 
                               ! (lda,*) and equivalently for B and C.
      integer :: ierr ! error code
      complex(kind=wp) :: a(:,:)
      complex(kind=wp) :: b(:,:)
      complex(kind=wp) :: c(:,:)
      complex(kind=wp), optional :: coeff_A,coeff_C
      logical, optional :: transpose_A, transpose_B
      complex(kind=wp) :: coeff_A_local,coeff_C_local
      character(len=1) :: transpose_A_local, transpose_B_local


      intent(in) a,b, transpose_A, transpose_B,coeff_A,coeff_C
      intent(inout) c

! Default values
      transpose_A_local='n'
      transpose_B_local='n'
      coeff_A_local = (1.0_wp, 0.0_wp)
      coeff_C_local = (0.0_wp, 0.0_wp)
      k = size(A,1)
      lda = max(1,size(a,1))
      ldb = max(1,size(b,1))
      ldc = max(1,size(c,1))
      m = size(C,1)
      n = size(C,2)

      if (present(transpose_A)) then
        if (transpose_A) then
          transpose_A_local='y'
          k = size(A,2)
        endif
      endif

      if (present(transpose_B)) then
        if (transpose_B) then
          transpose_B_local='y'
        endif
      endif

      if (present(coeff_A)) then
        coeff_A_local = coeff_A
      endif

      if (present(coeff_C)) then
        coeff_C_local = coeff_C
      endif

      ierr = 0
     
! call the Level 3 blas routine, assuming that we don't need to 
! transpose either input matrix (hence the two 'n's), and that we 
! simply want to calculate c = a*b  i.e. c = 1.0_wp*a*b + 0.0_wp*c
!
! dgemm will also do some error checking.  For example, it will stop 
! if any of m,n,k are < 0

      call zgemm(transpose_A_local,transpose_B_local,m,n,k,coeff_A_local,a,lda,b,ldb,coeff_C_local,c,ldc)

      end subroutine maths_zmatrix_multiply_blas95


!*==tcoef.spg  processed by SPAG 6.56Rc at 13:54 on  8 Oct 2010
      SUBROUTINE MATHS_TCOEF(A,B,NAO,NMO)
!***********************************************************************
!
!     Utility to transpose a matrix. A is transposed and written to B
!     leaving A unchanged.
!
!***********************************************************************
      IMPLICIT NONE
!
!*** Start of declarations rewritten by SPAG
!
! Dummy arguments
!
      INTEGER :: NAO, NMO
      REAL(KIND=wp), DIMENSION(NAO,NMO) :: A
      REAL(KIND=wp), DIMENSION(NMO,NAO) :: B
      INTENT (IN) A, NAO, NMO
      INTENT (OUT) B
!
! Local variables
!
      INTEGER :: I, J
!
!*** End of declarations rewritten by SPAG
!
      DO I=1, NMO
         DO J=1, NAO
            B(I,J)=A(J,I)
         END DO
      END DO
!
      RETURN
!
      END SUBROUTINE MATHS_TCOEF

      SUBROUTINE MATHS_CVECA_PLUS_VECB(N,X,VECA,INCA,VECB,INCB)
! Utility to interface to the BLAS routine DAXPY, which evaluates a 
! constant (X) times a vector (VECA) plus a vector (VECB).
      IMPLICIT NONE

      INTEGER :: N, INCA, INCB
      REAL(KIND=wp) :: X
      REAL(KIND=wp), DIMENSION(*) :: VECA, VECB

      CALL DAXPY(N,X,VECA,INCA,VECB,INCB)

      END SUBROUTINE MATHS_CVECA_PLUS_VECB

!*==mxm.spg  processed by SPAG 6.56Rc at 13:54 on  8 Oct 2010
      SUBROUTINE MATHS_MXM(A,NAR,B,NAC,C,NBC)
      USE consts, ONLY : ZERO=>XZERO
      IMPLICIT NONE
!
!*** Start of declarations rewritten by SPAG
!
! Dummy arguments
!
      INTEGER :: NAC, NAR, NBC
      REAL(KIND=wp), DIMENSION(NAR,*) :: A, C
      REAL(KIND=wp), DIMENSION(NAC,*) :: B
      INTENT (IN) A, B, NAC, NAR, NBC
      INTENT (INOUT) C
!
! Local variables
!
      INTEGER :: I, J, K
!
!*** End of declarations rewritten by SPAG
!
!**********************************************************************
!
!    Routine to emulate CRAY SCILIB routine, MXM
!
!**********************************************************************
!
      DO J=1, NBC
         DO I=1, NAR
            C(I,J)=zero
            DO K=1, NAC
               C(I,J)=C(I,J)+A(I,K)*B(K,J)
            END DO
         END DO
      END DO
!
      RETURN
      END SUBROUTINE MATHS_MXM

end module maths
